/*
  Software for OBI Plug with WiFi ESP Chip
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  Wifi setup code is based on the code of nikant at https://github.com/nikant/espWiFi2eeprom
  Also code is based on https://github.com/esp8266/Arduino
  modified by TomT,sissiwup
  contact : info@tomz.ws
*/
/*
  Default SoftAP IP is : 192.168.4.1
  Default SoftAP Port is : 80
  Default SoftAP Password is : esp8266-1
  COM Port : 115200 Baud 8N1
*/
/*
    ########## Pins ############
    GPIO4 : LED BLUE
    GPIO5 : Relay OFF
    GPIO12 : Relay ON
    GPIO14 : Push Button
*/
/*
   ################ READ ME BEFORE ######################
    Please use the following settings to flash the program to your OBI Plug:
 *  * ##### programming settings for ARDUINO #####
    Board : generic ESP8266 Module
    Flash Mode : DIO
    Flash Size : 1M (64k SPIFFS)
    Debug Port : disabled
    Debug Level : none
    Reset Method : ck
    Crystal frequency : 26 MHz
    Flash frequency : 40 MHz
    CPU frequency : 80 MHz
    Upload Speed : 115200
    After flash is done, you can remove the Plug from your computer USB UART Connection.
    Plug into a mains power wall plug.
    Look for a Wifi Access point calles somewhat like ESP_XYZ
    Make a connection to this Access Point and type into your browser
    192.168.4.1.
    You will see a page where you can choose your Homenetwork.
    Type in the Name and SSID Password and after that reboot the ESP.
    Then it will login into your local network and you now need to find the correct IP Adress.
    Type in the IP Adress and a page will be shown, wehere you can configure and Switch ON / OFF the Obi Plug.
    In future i will add functions.
    The OTA Over the air Update is integrated and you can updtae to a new Firmware version without opening the Plug again.
    DISCLAIMER / WARRANTY:
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
    TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING
    THE SOFTWARE BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT,
    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
    OR OTHER DEALINGS IN THE SOFTWARE.
    USE ANYTHING HERE AT YOUR OWN RISK....
   #########################################################
*/
/*
    #########################################################
    The Software is tested under Arduino Version 1.8.5
    and ESP8266 CORE Version 2.41
*/
/*
    #########################################################
    VERSION HISTORY :
    1.0.4
    - added to try to reconnect to a network after 3 Minutes running AP Point (fix for slow WLAN Startup after MAINS Power lost of ROUTER)
    - CCU CUXD Implementation
    1.0.6 
    - dynamic Site
    - Watch-Dog
    1.0.7
    - ntp-Server-Access
    1.0.8
    - devname
    1.0.9
    - IP-Adresse
    1.1.0
    - Bugfixes
    1.1.1
    - Performance / Webserver split
    1.1.2
    - not published
    1.1.3
    - Add SPIFFS
    1.1.4
    - VPN-Access optimized
    1.1.5
    - MQTT + VPN working
    1.1.6
    - Code cleared
*/
/*
   DON'T CHANGE ANYTHING BELOW!
   EVERYTHING WILL BE CONBFIGURED LATER WITH YOUR WEBBROWSER
*/
#include "version.h"
#include "webstatic.h"
#include <core_version.h>
#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <WiFiClient.h>
#include <ESP8266mDNS.h>
#include <ESP8266HTTPUpdateServer.h>
#include <ESP8266HTTPClient.h>
#include <Time.h>
#include <TimeLib.h>
#include <NtpClientLib.h>
#include <Ticker.h>
#include <ESP8266Ping.h>
#include <EEPROM.h>
#include "FS.h"
// #include <ESP8266WiFiMulti.h>
#include <PubSubClient.h>
// ESP8266WiFiMulti wifiMulti;
boolean connectioWasAlive = true;
extern "C"
{
  bool system_update_cpu_freq(uint8 freq);
}

extern "C"
{
  uint8 system_get_cpu_freq();
}

uint fcpufreq = 80;
int8_t timeZone = 1;
int8_t minutesTimeZone = 0;
FlashMode_t ideMode = ESP.getFlashChipMode();

FSInfo fs_info;
//System  **** DONT CHANGE THIS or system may be fail to run ****

String wwwInfos = "";   //will be filled in later with actual data
String wwwStatus1 = ""; //takes the last Status message
String wwwStatus2 = ""; //takes the last Status message

String wwwCommandHeader = "";
String wwwHeader1;
String wwwHeader2;
String wwwFooter;

int first_Startup = 0; //Do we start at the first time ever?

const String wwwGoBack = "<br><a href=\"/\" target=\"_self\">OK - GO BACK</a>"; //holds the code for the back to root button
int eeprom_Function = 0;                                                        //define what kind of data to be written to the EEPROM
int stateONPowerON = 0;                                                         //define the Realy state on Power ON 0=OFF 1= ON
int iNTPServer = 0;

String iDevname = "ESP8266-Wall-Plug-OBI";

IPAddress fIPown(192, 168, 1, 77);
IPAddress fIPgateway(192, 168, 1, 100);
IPAddress fIPsubnet(255, 255, 255, 0);
int fIPstatus = 0;

int ajaxon = 1; //(0=Off)

String wifi_SSID = "";
String wifi_PWD = "";

unsigned long timeNow = 0; //timestamp for timeroperations / msec.

unsigned long DeltaKeepWIFIAlive = 1800 * 1000; //Reconnect Delta Time / msec. (default 15 min.)
unsigned long LastKeepWIFIAlive = 0;            //timestamp of last Reconnect / msec.
int enableReconnect = 1;                        //switch Wifi.reconnect function on or off 0=OFF 1=ON

unsigned long DeltaTimeButton = 250; //Button Check Delta Time/ msec.
unsigned long LastTimeButton = 0;    //timestamp of last Button Check / msec.
boolean ButtonIsBlocked = false;     //We dont want a blinking switch when button wasn't released

int pingTest = 1;
long badPingCount = 0;
unsigned long LastPingTime = 0;
unsigned long DeltaPingTime = 30000;

String relStatus = "OFF"; //holds the Relay Status

unsigned long iDelayON = 0;
unsigned long iDelayOFF = 0;
unsigned long iONDelayOFF = 0;
unsigned long iOFFDelayON = 0;
unsigned long iAutoONOFF = 0;
unsigned long dAutoONOFF = 0;
unsigned long iDelayWatchDogOn = 0;
unsigned long iDelayWatchDogOff = 0;
unsigned long HardwareDogOnTime = 600 * 1000;
unsigned long HardwareDogOffTime = 10 * 1000;
boolean enableDelay = false;
int handle_Delay = 0;
//ioBroker
unsigned long DeltaSendState = 600 * 1000; //define the time for StatusMessage to ioBroker / msec.
unsigned long LastSendState = 0;           //timestamp of last StatusMessage / msec.
char host_ioBrokerIP[16] = "192.168.1.13"; //IP Adress ioBroker
int host_ioBrokerPORT = 8087;              //Portadress simple API Adapter
char IOB_Datapoint_Path[80] = "esp8266.0"; //Name of ioBroker Datapoint path
char IOB_Datapoint_Sub[80] = "DOSE.1";     //Name of ioBroker Subfolder
char IOB_Datapoint_State[80] = "STATE";    //Name of ioBroker Status datapoint
char IOB_Datapoint_Message[80] = "INFO";   //Name of ioBroker Message datapoint
String IOB_State = "";                     //ioBroker State Informations
String IOB_Message = "";                   //ioBroker String to be sent to Message Datapoint
int IOB_active = 0;                        //is ioBroker State Report active? 0=no 1=yes
//CCU
int CCU_CuxD = 1;                         //Send state to CuxD enabled 0=disabled 1=enabled
char CCU_CuxDAdress[38] = "CUX2801003:?"; //Name of CCU CuxD Device Adress incl. Channel
char host_CCUIP[16] = "192.168.1.11";     //IP Adress CCU2
//MQTT
int MQTT_active = 0;
char host_MQTT[16] = "192.168.1.13";
char path_MQTT[80] = "/SW-50/5";
unsigned long DeltaLastMQTTtime = 1000 * 30; // reconnect alle 30 Sekunden
unsigned long LastMQTTtime = 0;
//WEB-Server/Client
ESP8266WebServer server(80);
ESP8266HTTPUpdateServer httpUpdater;
WiFiClient espClient;
//MQTT-Client
PubSubClient pubClient(espClient);
//Blue LED
Ticker blinker;

// We create a Timestamp
String timeStamp()
{
  char ctimeStamp[10];
  unsigned long amillis = millis();
  unsigned long timeDay = amillis / 86400000;
  amillis = amillis % 86400000;
  unsigned long timeHour = amillis / 3600000;
  amillis = amillis % 3600000;
  unsigned long timeMinute = amillis / 60000;
  amillis = amillis % 60000;
  unsigned long timeSecond = amillis / 1000;
  sprintf(ctimeStamp, "%lu:%02lu:%02lu:%02lu", timeDay, timeHour, timeMinute, timeSecond);
  return ctimeStamp;
}

String gerDate(String inp)
{
  inp.replace("/", ".");
  return inp;
}

void buildwwwStrings()
{
  wwwHeader1 = String(F("<!DOCTYPE HTML>\n"
                        "<html><head><meta content=\"text/html;charset=utf-8\"><title>")) +
               iDevname + String(F("</title>\n"
                                   "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">"
                                   "<style type=\"text/css\">body {text-align:center;font-family: sans-serif;background-color: #1B2E6C;color: #AAAAAA;font-size:0.9em;} a:link, a:visited, a:hover, a:active {color:#fff;}</style>"));
  wwwHeader2 = String(F("</head>\n"
                        "<body>\n"
                        "<center>\n"
                        "<table border=1 width=\"80%\" cellpadding=\"6\"><tr><td colspan=2>"
                        "<h2><span id=\"devName\">")) +
               iDevname + String(F("</span></h2>"
                                   "</td></tr>\n\n"));
  wwwFooter = String(F("</table><br><hr>\n"
                       "<p style=\"font-size:0.8em\" a:link, a:visited, a:hover, a:active {color:#fff;>2018 by TomT/sissiwup - - - Version ")) +
              VERSION + "(" + iRomversion + ") Adruino : " + ESP.getCoreVersion() + String(F(" - - - Info : <a href=\"https://forum.iobroker.net/viewtopic.php?f=35&t=12125\" target=_new>https://forum.iobroker.net</a><br>"
                                                                                             "AP for WIFI settings is found at 192.168.4.1, if the ESP is not connected to a network. "
                                                                                             "If ESP fails to connect to a network, it will start an AP after 60 sec. timeout<br></p></center>\n"
                                                                                             "</body></html>"));
}

void handle_root()
{
  wwwInfos = String(F("<tr><td><b>ESP8266</b> <span id=\"debstat\"></span><br><br>own IP : ")) +
             WiFi.localIP().toString() + " (Mac:" + WiFi.macAddress() + ")<br>GATEWAY : " +
             WiFi.SSID() + "<br>Gateway IP : " + WiFi.gatewayIP().toString() +
             "<br>Gateway Mask : " + WiFi.subnetMask().toString() + " Ping : " + String(pingTest) +
             String(F(" (0=off/1=on) <b><a href=/ping?=1>ON</a> <a href=/ping?=0>OFF</a></b><br>FIX-IP-Status: <span id=\"ipStatus\">")) +
             fIPstatus + String(F("</span> (0=off/1=on) <b><a href=/fIPstatus?=1>ON</a> <a href=/fIPstatus?=0>OFF</a> (ON->OFF:reboot)</b>"
                                  "<br>FIX-IP-Address: <span id=\"ipAddr\">")) +
             fIPown.toString() + String(F("</span><br>FIX-IP-Gateway: <span id=\"ipGateway\">")) +
             fIPgateway.toString() + String(F("</span><br>FIX-IP-Subnet:  <span id=\"ipSubnet\">")) +
             fIPsubnet.toString() + String(F("</span><br>NTP-Server: <span id=\"ntpServer\">")) +
             NTP.getNtpServerName() + String(F("</span><br>RSSI :  <span id=\"wifiRssi\">")) +
             WiFi.RSSI() + String(F("</span> dBm CPU : <span id=\"cpufreq\">")) +
             String(system_get_cpu_freq()) +
             String(F("</span> <b><a href=/cpufreq?=80>80</a> <a href=/cpufreq?=160>160</a></b>"
                      "<br>FlashSize PGM/REAL(Heap):mode: ")) +
             ESP.getFlashChipSize() + "/" + ESP.getFlashChipRealSize() +
             "(<span id=\"freeHeap\">" + String(ESP.getFreeHeap()) + "</span>):" + (ideMode == FM_QIO ? "QIO" : ideMode == FM_QOUT ? "QOUT" : ideMode == FM_DIO ? "DIO" : ideMode == FM_DOUT ? "DOUT" : "UNKNOWN") +
             "<br>SPIFFS total/used : " + fs_info.totalBytes + " / " + fs_info.usedBytes + String(F("<br>Time since start : <span id=\"time\">")) +
             timeStamp() + "</span> (" +
             gerDate(NTP.getTimeDateString(NTP.getFirstSync())) + ")<br>";
  if (ajaxon == 1)
  {
    wwwInfos = wwwInfos +
               String(F("Response : <span id=\"reqtime\">?</span>ms. Avg : <span id=\"avgtime\">?</span>ms Blocked : <span id=\"blocker\">-</span>sec."));
  };
  wwwInfos = wwwInfos + String(F("<br>Actual Time : <span id=\"akttime\">")) + gerDate(NTP.getTimeDateString()) +
             String(F("</span><br>Time to Event : <span id=\"delayTimer\">OFF</span><br>"
                      "<p style=\"font-size:1.5em;color:white\"><b>SWITCH STATUS : <span id=\"relStatus\">")) +
             relStatus +
             String(F("</span></b><span id=\"delayStatus\">/")) + getDelayStatus() +
             String(F("</span></p><b>SWITCH: <a style=\"color:lightgreen\" href=/ON>ON</a> <a style=\"color:red\" href=/OFF>OFF</a></b></td>"
                      "<td><b>SETTINGS :</b><br>PowerONState : ")) +
             stateONPowerON + String(F(" <b><a href=/powerONState?=1>ON</a> <a href=/powerONState?=0>OFF</a> <a href=/powerONState?=2>Last state</a></b><br>"
                                       "Wifi-Reconnect enabled : ")) +
             enableReconnect + String(F(" <b><a href=/reConnect?=1>ON</a> <a href=/reConnect?=0>OFF</a></b><br>"
                                        "Wifi.Reconnect time : <span id=\"reconnectStatus\"></span>")) +
             NTP.getTimeStr(DeltaKeepWIFIAlive / 1000) +
             String(F(" <b><a href=/keepalive?=3600>1h</a> <a href=/keepalive?=21600>6h</a> <a href=/keepalive?=43200>12h</a> <a href=/keepalive?=86400>24h</a></b><br>"
                      "Ajax : ")) +
             ajaxon + String(F(" <b><a href=/ajax?=1>ON</a> <a href=/ajax?=0>OFF</a></b><br><br><b>ioBroker : </b><br>Active : ")) +
             String(IOB_active) + String(F("<b> <a href=/ioBroker?active=1>ON</a> <a href=/ioBroker?active=0>OFF</a></b><br>IP : ")) +
             String(host_ioBrokerIP) + "<br>SimpleAPI Port : " +
             host_ioBrokerPORT + "<br>Datapoint : " + IOB_Datapoint_Path + "(.)" + IOB_Datapoint_Sub + "<br>State : " +
             IOB_Datapoint_State + " / Message : " + IOB_Datapoint_Message + "<br>State Time Report : " +
             NTP.getTimeStr(DeltaSendState / 1000) + " <br><br><b>CCU : </b><br>IP : " +
             String(host_CCUIP) + "<br>" + " CuxD Device : " + String(CCU_CuxDAdress) + "<br> CuxD enabled : " +
             CCU_CuxD + String(F(" <b><a href=/CCU?cuxd=1>ON</a> <a href=/CCU?cuxd=0>OFF</a></b><br><br><b>MQTT : </b><br>IP : ")) +
             String(host_MQTT) + "<br>" + " Path : " + String(path_MQTT) + "<br> MQTT enabled : " +
             MQTT_active + String(F(" <b><a href=/mqtt?active=1>ON</a> <a href=/mqtt?active=0>OFF</a></b><br>"
                                    "<br><b>WatchDog : </b><br>On-Time : ")) +
             NTP.getTimeStr(HardwareDogOnTime / 1000) + " / Off-Time : " +
             NTP.getTimeStr(HardwareDogOffTime / 1000) +
             String(F("<br></td></tr><tr><td>Last States:<br></td><td>Last Service:</td></tr><tr><td><span id=\"wwwStatus1\">")) +
             wwwStatus1 + String(F("</span></td><td><span id=\"wwwStatus2\">")) +
             wwwStatus2 + String(F("</span></td></tr>"));
  //           + "<br>SSID=" + wifi_SSID + "/" + wifi_PWD + ""
  //           +"</td></tr>";
  //wh  server.send(200, "text/html", wwwHeader +  wwwInfos + wwwCommandlist + wwwFooter);
  Serial.println("Start-Webpage");
  sendWebPageChunkedBegin();
  sendWebPageChunkedData(wwwHeader1);
  if (ajaxon == 1)
  {
    sendWebPageChunkedData(wwwScript);
  }
  sendWebPageChunkedData(wwwHeader2);
  sendWebPageChunkedData(wwwInfos);
  sendWebPageChunkedData(wwwCommandlist);
  sendWebPageChunkedData(wwwFooter);
  sendWebPageChunkedEnd();
  Serial.println("End-Webpage");
  //  server.send(200, "text/html", wwwHeader1 + FPSTR(wwwScript) + wwwHeader2 +  wwwInfos + FPSTR(wwwCommandlist) + wwwFooter);
  delay(100);
}
void setup()
{
  Serial.begin(115200, SERIAL_8N1, SERIAL_TX_ONLY);
#ifdef DEBUG_ESP_PORT
  Serial.setDebugOutput(true);
#endif
  // Serial.begin (115200);
  EEPROM.begin(1024);
  /*
    ########## Pins ############
    GPIO4 : LED BLUE
    GPIO5 : Relay OFF
    GPIO12 : Relay ON
    GPIO14 : Push Button
*/
  //set the Pins to input or output mode
  pinMode(14, INPUT);  // BUTTON
  pinMode(4, OUTPUT);  // LED BLUE
  pinMode(5, OUTPUT);  // RELAY OFF
  pinMode(12, OUTPUT); // RELAY ON
  digitalWrite(5, 1);
  digitalWrite(12, 1);
  startBlink(0.5);
  // EEProm reading / Default-Values
  readFirstStart();
  if (first_Startup == 0)
  {
    delay(100);
    firstStart();
  }
  readVersion();
  writeInitValues();
  readEEPROM();
  spiffInit();
  system_update_cpu_freq(fcpufreq);
  // All Values initialized
  buildwwwStrings();
  char tmp[32];
  iDevname.toCharArray(tmp, 32);
  WiFi.persistent(false); // set WiFi credentials persistance to false to prevent eeprom corruption
  WiFi.setOutputPower(20.5);
  espNKWiFiconnect(wifi_SSID, wifi_PWD);
  startBlink(0.5);
  // from here Wifi is connected
  WiFi.hostname(tmp);
  if (MDNS.begin(tmp))
  {
    build_wwwStatus2(F("MDNS server started"));
  }
  WiFi.setAutoReconnect(true);
  WiFi.setAutoConnect(true);
  if (fIPstatus == 1)
  {
    WiFi.config(fIPown, fIPgateway, fIPsubnet);
  }
  Serial.println("IP address: " + WiFi.localIP().toString());
  server.on("/", handle_root);
  server.on("/updateInfo", handle_update);               //wh
  server.on("/watchDogRestart", handle_watchDog);        //wh
  server.on("/watchDogOnTime", handle_watchDogOnTime);   //wh
  server.on("/watchDogOffTime", handle_watchDogOffTime); //wh
  server.on("/toggle", handle_toggle);                   //wh
  server.on("/getState", handle_getState);               //wh
  server.on("/ntp", handle_ntpServer);
  server.on("/devname", handle_devName);
  server.on("/fIPstatus", handle_IPstatus);
  server.on("/fIP", handle_IPown);
  server.on("/fIPgateway", handle_IPgateway);
  server.on("/fIPsubnet", handle_IPsubnet);
  server.on("/cpufreq", handle_cpufreq);
  server.on("/ajax", handle_ajax);
  server.on("/ping", handle_ping);
  server.on("/reboot", handle_APrestart);
  server.on("/erasessid", handle_clearAPeeprom);
  server.on("/eraseEEPROM", handle_clearAPalleeprom);
  server.on("/powerONState", handle_PowerONState);
  server.on("/reConnect", handle_Reconnect);
  server.on("/keepalive", handle_KeepAlive);
  server.on("/ioBroker", handle_ioBroker); //ioBroker Handle
  server.on("/ON", switch_ON);             //ON
  server.on("/OFF", switch_OFF);           //OFF
  server.on("/on", switch_ON);             //ON
  server.on("/off", switch_OFF);           //OFF
  server.on("/delayON", delay_ON);         //delayed ON
  server.on("/delayOFF", delay_OFF);       //delayed OFF
  server.on("/ONdelayOFF", ON_delay_OFF);  //direct ON and delayed OFF
  server.on("/OFFdelayON", OFF_delay_ON);  //direct OFF and delayed ON
  server.on("/AutoONOFF", AUTO_ON_OFF);    //repeated AUTO ON/OFF
  server.on("/CCU", handle_CCU);           //CCU Handle
  server.on("/mqtt", handle_mqtt);         //MQTT Handle
  //Send command ON
  server.on("/CommandON", command_ON);
  //Send command OFF
  server.on("/CommandOFF", command_OFF);
  server.onNotFound([]() {
    String message = "File Not Found\n\n";
    message += "URI: ";
    message += server.uri();
    message += "\nMethod: ";
    message += (server.method() == HTTP_GET) ? "GET" : "POST";
    message += "\nArguments: ";
    message += server.args();
    message += "\n";
    for (int i = 0; i < server.args(); i++)
    {
      message += " " + server.argName(i) + ": " + server.arg(i) + "\n";
    }
    server.send(404, "text/plain", message);
  });
  httpUpdater.setup(&server);
  server.begin();
  Serial.println(F("HTTP server started"));
  //wwwStatus = timeStamp() + " HTTP server started";
  MDNS.addService("http", "tcp", 80);
  timeNow = millis();
  LastKeepWIFIAlive = timeNow;
  LastTimeButton = timeNow;
  LastPingTime = timeNow;
  setInitialSwitchState();
  createCommandHeader();
  if (iNTPServer == 1)
  {
    IPAddress gate = fIPgateway;
    if (fIPstatus == 0)
    {
      gate = WiFi.gatewayIP();
    }
    NTP.begin(gate.toString(), timeZone, true, minutesTimeZone);
    build_wwwStatus2(F("NTP connection to Gateway"));
  }
  if (iNTPServer != 1)
  {
    NTP.begin("pool.ntp.org", timeZone, true, minutesTimeZone);
    build_wwwStatus2(F("NTP connection to pool.ntp.org"));
  }
  NTP.setInterval(63);
  build_wwwStatus2(F("HTTP server started"));
  startMQTTServer();
  stopBlink();
  delay(100);
  startCountBlink(0.5, 10);
}
void loop()
{
  timeNow = millis();
  server.handleClient();
  delay(10);
  monitorWiFi();
  if (MQTT_active == 1)
  {
    if (!pubClient.connected())
    {
      if (timeNow - LastMQTTtime > DeltaLastMQTTtime)
      {
        reconnect();
        LastMQTTtime = timeNow;
      }
    }
    pubClient.loop();
  }
  // Should that be a fix of Connection fails to the webserver?
  // we give it a try
  // see https://github.com/esp8266/Arduino/issues/2330
  //or use ping script that pings the ESP every 30 sec. (CuxD Ping device)
  // we test with 2 minutes
  // 30sec Test was ok
  if (timeNow - LastPingTime > DeltaPingTime && pingTest == 1)
  {
    IPAddress gate = fIPgateway;
    if (fIPstatus == 0)
    {
      gate = WiFi.gatewayIP();
    }
    if (!Ping.ping(gate, 1))
    {
      build_wwwStatus2(F("Gateway not reach"));
      badPingCount++;
    }
    else
    {
      badPingCount = 0;
    }
    if (CCU_CuxD == 1)
    {
      IPAddress ccuip;
      ccuip.fromString(String(host_CCUIP));
      if (!Ping.ping(ccuip, 1))
      {
        build_wwwStatus2(F("CCU not reach"));
      }
    }
    if (IOB_active == 1)
    {
      IPAddress iobip;
      iobip.fromString(String(host_ioBrokerIP));
      if (!Ping.ping(iobip, 1))
      {
        build_wwwStatus2(F("IOB not reach"));
      }
    }
    LastPingTime = timeNow;
  }
  if (enableReconnect == 1)
  {
    if (timeNow - LastKeepWIFIAlive > DeltaKeepWIFIAlive)
    {
      WiFi.reconnect();
      LastKeepWIFIAlive = timeNow;
      startCountBlink(0.5, 20);
      delay(10);
      build_wwwStatus2("Wifi.Reconnect Time done");
    }
  }
  if (badPingCount > 24)
  { // 12 * 2 * 5000ms = 2 Min
    badPingCount = 0;
    WiFi.reconnect();
    startCountBlink(0.5, 10);
    delay(10);
    build_wwwStatus2("Wifi.Reconnect bad Ping Gateway done");
  }
  // Check for a button press once a sec.
  if (timeNow - LastTimeButton > DeltaTimeButton)
  {
    checkButton();
    LastTimeButton = timeNow;
  }
  checkDelay();
  // timeLoop for StatusReports to IOB
  if (timeNow - LastSendState > DeltaSendState)
  {
    LastSendState = timeNow;
    if (IOB_active != 0)
    {
      reportIOB(-1);
      delay(10);
      build_wwwStatus2("sent Status to ioBroker");
    }
  }
}
//check for a button press
void checkButton()
{
  int Button = 1;
  //debounce loop to prevent false positive pulses
  int ButtonPress = 0;
  for (int i = 0; i < 16; i++)
  {
    Button = digitalRead(14);
    if (Button == 0)
    {
      ButtonPress++;
    }
  }
  //to prevent ON/OFF loop we switch once and wait for the button to be released
  if ((ButtonPress >= 15) && (ButtonIsBlocked == false))
  {
    if (relStatus == "OFF")
    {
      //Serial.println("Button ON");
      switch_ON();
      ButtonIsBlocked = true;
    }
    else
    {
      //Serial.println("Button OFF");
      switch_OFF();
      ButtonIsBlocked = true;
    }
  }
  if (Button == 1)
  {
    ButtonIsBlocked = false;
  }
}
/* set the initial Switch State on Power ON
   needed on future Options configured by the user
*/
void setInitialSwitchState()
{
  // default should by that Relais is OFF
  if (stateONPowerON == 0)
  {
    switch_OFF();
  }
  if (stateONPowerON == 1)
  {
    switch_ON();
  }
  if (stateONPowerON == 2)
  {
    if (getLastState() == 0)
    {
      switch_OFF();
    }
    else
    {
      switch_ON();
    }
  }
}
//We want some valid values in the EEPROM after the first start
// So we gonna generate us some
void firstStart()
{
  if (first_Startup == 0)
  {
    Serial.println(F("#### FIRST STARTUP ####"));
    Serial.println(F("generating default values now !"));
    Serial.println(F("Please wait..."));
    //We generate the EEPROM-values in a loop and call each Function one after the other
    first_Startup = 1;
    for (int i = 1; i < 100; i++)
    {
      //Serial.println("EEPROM_function : " + String(eeprom_Function));
      eeprom_Function = i;
      writeEEPROM();
      delay(10);
    }
    eeprom_Function = 0;
    build_wwwStatus2(F("default values written to EEPROM"));
  }
}
void createCommandHeader()
{
  wwwCommandHeader = String(F("<!DOCTYPE HTML>\n"
                              "<html><head><meta content=\"text/html;charset=utf-8\"><title>ESP8266 Wall Plug OBI</title>\n"
                              "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">"
                              "<meta http-equiv=\"refresh\" content=\"1; URL=http://")) +
                     WiFi.localIP().toString() +
                     String(F("/\">"
                              "<style type=\"text/css\">body {text-align:center;font-family: sans-serif;background-color: #1B2E6C;color: #AAAAAA;font-size:0.9em;} a:link, a:visited, a:hover, a:active {color:#fff;}</style>"
                              "</head>\n"
                              "<body>\n"
                              "<center>\n"));
}